package net.locusworks.s3sync.client;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import javax.annotation.PostConstruct;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import com.amazonaws.AmazonClientException;
import com.amazonaws.AmazonServiceException;
import com.amazonaws.regions.Regions;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.Bucket;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectSummary;
import com.amazonaws.services.s3.transfer.Download;
import com.amazonaws.services.s3.transfer.TransferManager;
import com.amazonaws.services.s3.transfer.TransferManagerBuilder;
import com.amazonaws.services.s3.transfer.Upload;
import com.jcraft.jsch.JSchException;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;
import net.locusworks.s3sync.conf.ConfigurationManager;
import net.locusworks.s3sync.conf.TransferType;
import net.locusworks.s3sync.database.entities.FileInfo;
import net.locusworks.s3sync.scp.AbstractSshMessage;
import net.locusworks.s3sync.scp.LocalFileMessage;
import net.locusworks.s3sync.scp.ScpFromMessage;

@Component
public class S3Client {

  private ApplicationLogger logger = ApplicationLoggerFactory.getLogger(S3Client.class);

  private AmazonS3 s3Client;

  @Autowired
  private ConfigurationManager conf;
  
  @Autowired
  private FileManager fileManager;
  
  public S3Client() {}
  
  @PostConstruct
  private void init() {
    String region = conf.getRegion();
    this.s3Client = AmazonS3ClientBuilder.standard().withRegion(Regions.fromName(region)).build();
    Bucket bucket = s3Client.listBuckets().stream()
        .filter(b -> b.getName().equals(conf.getBucketName())).findFirst().orElse(null);

    if (bucket == null) {
      logger.error("Unable to find bucket with name: %s%nExiting", conf.getBucketName());
      System.exit(-1);
    }
    logger.info("Found Bucket: %s", bucket);
  }
  
  public String getBucket() {
    return this.conf.getBucketName();
  }

  public void uploadFile(Path file, Path localFolder) {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    uploadFile(xferMgr, file, localFolder);
    xferMgr.shutdownNow(false);
  }

  public void uploadFile(TransferManager xferMgr, Path file, Path localFolder) {
    boolean xferMgrNull = xferMgr == null;
    xferMgr = !xferMgrNull ? xferMgr : TransferManagerBuilder.standard().withS3Client(s3Client).build();
    FileInfo fd = null;
    try {
      String key = getPath(file, localFolder);
      fd = fileManager.getFileDetail(file, key);
      if (fd.getUploaded()) return;
      logger.info("Uploading file: %s", file);
      Upload xfer = xferMgr.upload(getBucket(), key, file.toFile());
      xfer.waitForCompletion();
      fd.setUploaded(true);
      logger.info("Done uploading %s", file);
    } catch (AmazonClientException | InterruptedException | IOException e) {
      if (fd != null) {
        fd.setUploaded(false);
      }
      logger.error(e.getMessage());
    } finally {
      if (fd != null) {
        fileManager.addEntry(fd);
      }
      if (xferMgrNull) {
        xferMgr.shutdownNow(false);
      }
    }
  }
  
  public void downloadFile(String key, Path file) {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    downloadFile(xferMgr, key, file);
    xferMgr.shutdownNow(false);
  }

  public void downloadFile(TransferManager xferMgr, String key, Path file) {
    boolean xferMgrNull = xferMgr == null;
    xferMgr = !xferMgrNull ? xferMgr : TransferManagerBuilder.standard().withS3Client(s3Client).build();
    try {
      logger.info("Downloading file: %s", file);
      Download xfer = xferMgr.download(getBucket(), key, file.toFile());
      xfer.waitForCompletion();
      logger.info("Done downloading %s", file);
    } catch (AmazonClientException | InterruptedException e) {
      logger.error(e.getMessage());
    } finally {
      if (xferMgrNull) {
        xferMgr.shutdownNow(false);
      }
    }
  }

  public void syncFolder() throws IOException, JSchException {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    Path localFolder = conf.getLocalFolder();
    if (Files.notExists(localFolder)) {
      logger.info("Creating local folder: " + localFolder);
      Files.createDirectory(localFolder);
    }

    AbstractSshMessage aMsg = conf.getTransferType() == TransferType.REMOTE ? 
          new ScpFromMessage(true, ScpFromMessage.newSession(conf.getIdentity(), conf.getUser(), conf.getHost(), conf.getPort()), conf.getRemoteFolder(), localFolder, true) :
          new LocalFileMessage(localFolder);
    
    try(AbstractSshMessage msg = aMsg) {
      msg.fileCallback(p -> {
        if (Files.isRegularFile(p)) {
          uploadFile(xferMgr, p, localFolder);
          try {
            if (conf.deleteLocal())
              Files.delete(p);
          } catch (Exception ex) {
            logger.warn("Unable to delete local file %s: %s", p, ex.getMessage());
          }
        }
      });
      msg.execute();
    } catch (Exception e) {
      logger.error("Unable to load file Manager: " + e.getMessage());
    } 
    
    xferMgr.shutdownNow(false);
    if (conf.deleteLocal()) 
      Files.walk(localFolder)
        .sorted(Comparator.reverseOrder())
        .map(Path::toFile)
        .forEach(File::delete);
  }
  
  public S3Object getObject(String bucketName, String key) {
    try {
      return s3Client.getObject(bucketName, key);
    } catch (AmazonServiceException ex) {}
    return null;
  }
  
  public Set<String> getFileList() {
    Set<String> fileList = new HashSet<String>();
    for(S3ObjectSummary os: s3Client.listObjectsV2(getBucket()).getObjectSummaries()) {
      fileList.add(os.getKey());
    }
    return fileList;
  }
  
  public void removeFiles(Set<String> s3Files) {
    for(String key : s3Files) {
      try {
        logger.info("Removing file: " + key);
        s3Client.deleteObject(getBucket(), key);
      } catch (AmazonServiceException ex) {
        logger.warn(String.format("Unable to delete %s: %s", key, ex.getMessage()), ex);
      }
    }
  }
 
  private String getPath(Path file, Path localFolder) {
    if (file.getParent() == null) return file.getFileName().toString();
    Path relative = localFolder.relativize(file);
    return relative.toString().replace("\\", "/");
  }
  
  public void close() throws Exception {
    if (s3Client != null) {
      s3Client.shutdown();
    }
  }

}
