package net.locusworks.s3sync.database;

import java.util.HashSet;
import java.util.Properties;

import javax.sql.DataSource;

import org.flywaydb.core.Flyway;
import org.flywaydb.core.api.configuration.FluentConfiguration;
import org.flywaydb.core.api.logging.Log;
import org.flywaydb.core.api.logging.LogCreator;
import org.flywaydb.core.api.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.DependsOn;
import org.springframework.context.annotation.Primary;
import org.springframework.data.jpa.repository.config.EnableJpaRepositories;
import org.springframework.orm.hibernate5.LocalSessionFactoryBean;
import org.springframework.orm.jpa.JpaTransactionManager;
import org.springframework.orm.jpa.JpaVendorAdapter;
import org.springframework.orm.jpa.LocalContainerEntityManagerFactoryBean;
import org.springframework.orm.jpa.vendor.Database;
import org.springframework.orm.jpa.vendor.HibernateJpaVendorAdapter;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.annotation.EnableTransactionManagement;

import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;
import net.locusworks.logger.ApplicationLoggerInitializer;
import net.locusworks.logger.LogLevel;
import net.locusworks.s3sync.conf.ConfigurationManager;

@Configuration(value="net.locusworks.s3sync.database.S3SyncBeanConfiguration")
@EnableTransactionManagement
@EnableJpaRepositories(
    basePackages = {"net.locusworks.s3sync.database.repos"},
    entityManagerFactoryRef = "s3syncEntityManagerFactory",
    transactionManagerRef = "s3syncTransactionManager"
)
public class S3SyncBeanConfiguration {
  
  public static final String PERSISTENCE_UNIT = "s3sync_pu";
  
  public static final HashSet<String> cacheNames = new HashSet<>();

  private static final String[] PACKAGES_TO_SCAN = {"net.locusworks.s3sync.database.entities", "net.locusworks.s3sync.database.repos"};

  private static final Properties hibernateProperties;
  static {
    hibernateProperties = new Properties();
    hibernateProperties.setProperty("hibernate.connection.zeroDateTimeBehavior", "convertToNull");
    hibernateProperties.setProperty("hibernate.dbcp.maxActive", "50");
    hibernateProperties.setProperty("hibernate.dbcp.maxIdle", "10");
    hibernateProperties.setProperty("hibernate.dbcp.maxWait", "5000");
    hibernateProperties.setProperty("hibernate.jdbc.batch_size property", "50");
    hibernateProperties.setProperty("hibernate.connection.charSet", "UTF-8");
    hibernateProperties.setProperty("hibernate.connection.characterEncoding", "UTF-8");
    hibernateProperties.setProperty("hibernate.connection.useUnicode", "true");
  }
  
  @Autowired
  private DataSource dataSource;

  /**
   * Create the entity manager factory bean used in the database connection
   * @return entityManagerFactoryBean
   */
  @Primary
  @Bean
  @DependsOn("flyway")
  public LocalContainerEntityManagerFactoryBean s3syncEntityManagerFactory() {
    LocalContainerEntityManagerFactoryBean lef = new LocalContainerEntityManagerFactoryBean();
    lef.setDataSource(dataSource);
    lef.setJpaVendorAdapter(s3syncJpaVendorAdapter());
    lef.setPackagesToScan(PACKAGES_TO_SCAN);
    lef.setJpaProperties(hibernateProperties);
    lef.setPersistenceUnitName(PERSISTENCE_UNIT);
    return lef;
  }

  /**
   * Create the session factory bean
   * @return sessionFactoryBean
   */
  @Bean
  public LocalSessionFactoryBean s3syncSessionFactory() {
    LocalSessionFactoryBean sessionBean = new LocalSessionFactoryBean();
    sessionBean.setDataSource(dataSource);
    sessionBean.setPackagesToScan(PACKAGES_TO_SCAN);
    sessionBean.setHibernateProperties(hibernateProperties);
    return sessionBean;
  }

  /**
   * Create the JPA Vendor adaptor bean that is used in the entity manager factory
   * @return jpaVendorAdaptor
   */
  @Primary
  @Bean
  public JpaVendorAdapter s3syncJpaVendorAdapter() {
    HibernateJpaVendorAdapter hibernateJpaVendorAdapter = new HibernateJpaVendorAdapter();
    hibernateJpaVendorAdapter.setShowSql(false);
    hibernateJpaVendorAdapter.setGenerateDdl(false);
    hibernateJpaVendorAdapter.setDatabase(Database.H2);
    return hibernateJpaVendorAdapter;
  }

  /**
   * Create the transaction manager bean 
   * @return transactionManager
   */
  @Primary
  @Bean
  public PlatformTransactionManager s3syncTransactionManager() {
    JpaTransactionManager manager = new JpaTransactionManager();
    manager.setEntityManagerFactory(s3syncEntityManagerFactory().getObject());
    return new JpaTransactionManager();
  }
  
  @Bean(name="flyway", initMethod="migrate")
  @DependsOn({"initializer"})
  public Flyway flyway() {
    LogFactory.setLogCreator(new FlywayLogCreator());
    FluentConfiguration fc = Flyway.configure();
    fc.schemas("s3sync");
    fc.table("_flyway_migration");
    fc.locations("database/migration");
    fc.outOfOrder(false);
    fc.dataSource(dataSource);
    
    return fc.load();
  }
  
  @Bean(name="initializer") 
  public Boolean initializer(ConfigurationManager confManager) {
    ApplicationLoggerFactory.init(new ApplicationLoggerInitializer() {

      @Override
      public LogLevel initialize() {
        return confManager.getLogLevel();
      }
    });
    return true;
  }
  
  private class FlywayLogCreator implements LogCreator {
    
    @Override
    public Log createLogger(Class<?> clazz) {
      return new FlywayLog(clazz);
    }
  }
  
  private class FlywayLog implements Log {
    
    private ApplicationLogger logger;
    
    public FlywayLog(Class<?> clazz) {
      logger = ApplicationLoggerFactory.getLogger(clazz);
    }

    @Override
    public boolean isDebugEnabled() {
      return logger.isDebugEnabled();
    }

    @Override
    public void debug(String message) {
      logger.debug(message);
    }

    @Override
    public void info(String message) {
      logger.info(message);
    }

    @Override
    public void warn(String message) {
      logger.warn(message);
    }

    @Override
    public void error(String message) {
      logger.error(message);
    }

    @Override
    public void error(String message, Exception e) {
      logger.error(message, e);
    }
    
  }
  
}
