package net.locusworks.s3sync.scp;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.text.NumberFormat;
import java.util.function.Consumer;
import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpProgressMonitor;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.LogLevel;

public abstract class AbstractSshMessage implements AutoCloseable {

  private ApplicationLogger logger;

  private static final double ONE_SECOND = 1000.0;

  private Session session;
  private final boolean verbose;
  private final boolean compressed;
  
  protected Consumer<Path> fileCallback;

  /**
   * Constructor for AbstractSshMessage
   * @param session the ssh session to use
   */
  public AbstractSshMessage(final Session session, ApplicationLogger logger) {
    this(false, session, logger);
  }

  /**
   * Constructor for AbstractSshMessage
   * @param verbose if true do verbose logging
   * @param session the ssh session to use
   * @since Ant 1.6.2
   */
  public AbstractSshMessage(final boolean verbose, final Session session, ApplicationLogger logger) {
    this(verbose, false, session, logger);
  }

  /**
   * Constructor for AbstractSshMessage
   * @param verbose if true do verbose logging
   * @param compressed if true use compression
   * @param session the ssh session to use
   * @since Ant 1.9.8
   */
  public AbstractSshMessage(boolean verbose, boolean compressed, Session session, ApplicationLogger logger) {
    this.verbose = verbose;
    this.compressed = compressed;
    this.session = session;
    this.logger = logger;
  }

  /**
   * Open an ssh channel.
   * @param command the command to use
   * @return the channel
   * @throws JSchException on error
   */
  protected Channel openExecChannel(final String command) throws JSchException {
    final ChannelExec channel = (ChannelExec) session.openChannel("exec");
    channel.setCommand(command);

    return channel;
  }

  /**
   * Open an ssh sftp channel.
   * @return the channel
   * @throws JSchException on error
   */
  protected ChannelSftp openSftpChannel() throws JSchException {
    return (ChannelSftp) session.openChannel("sftp");
  }

  /**
   * Send an ack.
   * @param out the output stream to use
   * @throws IOException on error
   */
  protected void sendAck(final OutputStream out) throws IOException {
    final byte[] buf = new byte[1];
    buf[0] = 0;
    out.write(buf);
    out.flush();
  }

  /**
   * Reads the response, throws a BuildException if the response
   * indicates an error.
   * @param in the input stream to use
   * @throws IOException on I/O error
   * @throws BuildException on other errors
   */
  protected void waitForAck(final InputStream in) throws IOException, JSchException {
    final int b = in.read();

    // b may be 0 for success,
    //          1 for error,
    //          2 for fatal error,

    if (b == -1) {
      // didn't receive any response
      throw new JSchException("No response from server");
    }
    if (b != 0) {
      final StringBuilder sb = new StringBuilder();

      int c = in.read();
      while (c > 0 && c != '\n') {
        sb.append((char) c);
        c = in.read();
      }

      if (b == 1) {
        throw new JSchException("server indicated an error: " + sb.toString());
      } else if (b == 2) {
        throw new JSchException("server indicated a fatal error: " + sb.toString());
      } else {
        throw new JSchException("unknown response, code " + b + " message: " + sb.toString());
      }
    }
  }

  /**
   * Carry out the transfer.
   * @throws IOException on I/O errors
   * @throws JSchException on ssh errors
   */
  public abstract void execute() throws IOException, JSchException;

  /**
   * Log a message to the log listener.
   * @param message the message to log
   */
  protected void log(final String message, LogLevel logLevel) {
    switch (logLevel) {
      case DEBUG:
        logger.debug(message);
        break;
      case ERROR:
      case FATAL:
        logger.error(message);
        break;
      case INFO:
        logger.info(message);
        break;
      case WARN:
        logger.warn(message);
        break;
      case TRACE:
        logger.trace(message);
        break;
      case OFF:
      default:
        logger.trace(message);
        break;
    }
  }

  /**
   * Log transfer stats to the log listener.
   * @param timeStarted the time started
   * @param timeEnded   the finishing time
   * @param totalLength the total length
   */
  protected void logStats(final long timeStarted, final long timeEnded, final long totalLength) {
    final double duration = (timeEnded - timeStarted) / ONE_SECOND;
    final NumberFormat format = NumberFormat.getNumberInstance();
    format.setMaximumFractionDigits(2);
    format.setMinimumFractionDigits(1);
    log("File transfer time: " + format.format(duration) + " Average Rate: " + format.format(totalLength / duration) + " B/s", LogLevel.DEBUG);
  }

  /**
   * Is the verbose attribute set.
   * @return true if the verbose attribute is set
   * @since Ant 1.6.2
   */
  protected final boolean getVerbose() {
    return verbose;
  }

  /**
   * Is the compressed attribute set.
   * @return true if the compressed attribute is set
   * @since Ant 1.9.8
   */
  protected final boolean getCompressed() {
    return compressed;
  }
  
  protected final Session getSession() {
    return session;
  }
  
  protected void connect() throws JSchException {
    if (session != null && !session.isConnected()) session.connect();
  }
  
  protected void disconnect() {
    if (session != null && session.isConnected()) session.disconnect();
  }
  
  public Consumer<Path> fileCallback(Consumer<Path> callback) {
    this.fileCallback = callback;
    return this.fileCallback;
  }
  
  @Override
  public void close() throws Exception {
    if (session != null && session.isConnected()) {
      session.disconnect();
      session = null;
    }
  }

  /**
   * Track progress every 10% if 100kb &lt; filesize &lt; 1Mb. For larger
   * files track progress for every percent transmitted.
   * @param filesize the size of the file been transmitted
   * @param totalLength the total transmission size
   * @param percentTransmitted the current percent transmitted
   * @return the percent that the file is of the total
   */
  protected final int trackProgress(final long filesize, final long totalLength,
      final int percentTransmitted) {

    // CheckStyle:MagicNumber OFF
    final int percent = (int) Math.round(Math.floor((totalLength / (double) filesize) * 100));

    if (percent > percentTransmitted) {
      if (filesize < 1048576) {
        if (percent % 10 == 0) {
          if (percent == 100) {
            System.out.println(" 100%");
          } else {
            System.out.print("*");
          }
        }
      } else {
        if (percent == 50) {
          System.out.println(" 50%");
        } else if (percent == 100) {
          System.out.println(" 100%");
        } else {
          System.out.print(".");
        }
      }
    }
    // CheckStyle:MagicNumber ON

    return percent;
  }

  private ProgressMonitor monitor = null;

  /**
   * Get the progress monitor.
   * @return the progress monitor.
   */
  protected SftpProgressMonitor getProgressMonitor() {
    if (monitor == null) {
      monitor = new ProgressMonitor();
    }
    return monitor;
  }

  private class ProgressMonitor implements SftpProgressMonitor {
    private long initFileSize = 0;
    private long totalLength = 0;
    private int percentTransmitted = 0;

    @Override
    public void init(final int op, final String src, final String dest, final long max) {
      initFileSize = max;
    }

    @Override
    public boolean count(final long len) {
      totalLength += len;
      percentTransmitted = trackProgress(initFileSize, totalLength, percentTransmitted);
      return true;
    }

    @Override
    public void end() {
    }

    @SuppressWarnings("unused")
    public long getTotalLength() {
      return totalLength;
    }
  }

}
